// file      : xsd/cxx/parser/validator.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/parser/validator.hxx>

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

#include <cxx/parser/elements.hxx>

#include <iostream>

using std::wcerr;

namespace CXX
{
  namespace Parser
  {
    namespace
    {
      class ValidationContext: public Context
      {
      public:
        ValidationContext (SemanticGraph::Schema& schema_,
                           CLI::Options const& options,
                           Boolean& valid_)
            : Context (std::wcerr, options),
              valid (valid_),
              schema (schema_)
        {
        }

      public:
        String
        xpath (SemanticGraph::Nameable& n)
        {
          if (n.is_a<SemanticGraph::Namespace> ())
            return L"<namespace-level>"; // There is a bug if you see this.

          if (n.named ())
          {
            SemanticGraph::Scope& scope (n.scope ());

            if (scope.is_a<SemanticGraph::Namespace> ())
              return n.name ();

            return xpath (scope) + L"/" + n.name ();
          }
          else
          {
            return L"(anonymous type for " +
              n.context ().get<String> ("instance-name") + L")";
          }
        }

      protected:
        ValidationContext (ValidationContext& c)
            :  Context (c),
               valid (c.valid),
               schema (c.schema)
        {
        }

      protected:
        Boolean& valid;
        SemanticGraph::Schema& schema;
      };

      //
      //
      struct Any : Traversal::Any, protected virtual ValidationContext
      {
        Any (ValidationContext& c)
            : ValidationContext (c)
        {
        }

        struct Element: Traversal::Element,
                        protected virtual ValidationContext
        {
          Element (ValidationContext& c, SemanticGraph::Any& any)
              : ValidationContext (c),
                any_ (any),
                ns_ (any.definition_namespace ().name ())
          {
          }

          virtual Void
          traverse (SemanticGraph::Element& e)
          {
	    if (skip (e)) return;

            using SemanticGraph::Any;

            Boolean q (e.qualified ());
            String ns (q ? e.namespace_ ().name () : "");

            for (Any::NamespaceIterator i (any_.namespace_begin ());
                 i != any_.namespace_end (); ++i)
            {
              Boolean failed (false);

              if (*i == L"##any")
              {
                failed = true;
              }
              else if (*i == L"##other")
              {
                if (ns_)
                {
                  // Note that here I assume that ##other does not
                  // include names without target namespace. This
                  // is not what the spec says but that seems to be
                  // the consensus.
                  //
                  failed = q && ns != ns_;
                }
                else
                {
                  // No target namespace.
                  //
                  failed = q && ns != L"";
                }
              }
              else if (*i == L"##local")
              {
                failed = !q || ns == L"";
              }
              else if (*i == L"##targetNamespace")
              {
                failed = (q && ns_ == ns) || (!q && ns_ == L"");
              }
              else
              {
                failed = q && *i == ns;
              }

              if (failed)
              {
                Any& a (any_);

                os << a.file () << ":" << a.line () << ":" << a.column ()
                   << ": error: namespace '" << *i << "' allows for element "
                   << "'" << e.name () << "'" << endl;

                os << e.file () << ":" << e.line () << ":" << e.column ()
                   << ": error: element '" << e.name () << "' is defined "
                   << "here" << endl;

                os << a.file () << ":" << a.line () << ":" << a.column ()
                   << ": error: such wildcards are not supported" << endl;

                valid = false;
              }
            }
          }

        private:
          SemanticGraph::Any& any_;
          String ns_;
        };

        struct Complex: Traversal::Complex
        {
          Complex ()
              : up_ (true), down_ (true)
          {
          }

          virtual Void
          post (Type& c)
          {
            // Go down the inheritance hierarchy.
            //
            if (down_)
            {
              Boolean up = up_;
              up_ = false;

              for (Type::InheritsIterator i (c.inherits_begin ());
                   i != c.inherits_end (); ++i)
              {
                dispatch (i->inheritee ());
              }

              up_ = up;
            }

            // Go up the inheritance hierarchy.
            //
            if (up_)
            {
              Boolean down = down_;
              down_ = false;

              for (Type::BegetsIterator i (c.begets_begin ());
                   i != c.begets_end (); ++i)
              {
                dispatch (i->inheritor ());
              }

              down_ = down;
            }
          }

        private:
          Boolean up_, down_;
        };

        virtual Void
        traverse (SemanticGraph::Any& a)
        {
          using SemanticGraph::Compositor;

          // Find our complex type.
          //
          Compositor* c (&a.contained_particle ().compositor ());

          while(!c->contained_compositor_p ())
            c = &c->contained_particle ().compositor ();

          SemanticGraph::Complex& type (
            dynamic_cast<SemanticGraph::Complex&> (
              c->contained_compositor ().container ()));

          Complex complex;
          Traversal::Names names;
          Element element (*this, a);

          complex >> names >> element;

          complex.dispatch (type);
        }
      };


      //
      //
      struct Traverser : Traversal::Schema,
                         Traversal::Complex,
                         Traversal::Type,
                         Traversal::Element,
                         Traversal::Attribute,
                         protected virtual ValidationContext
      {
        Traverser (ValidationContext& c)
            : ValidationContext (c),
              any_ (c)
        {
          *this >> sources_ >> *this;
          *this >> schema_names_ >> ns_ >> names_;

          names_ >> *this >> names_;

          // Any
          //
          *this >> contains_compositor_ >> compositor_ >> contains_particle_;
          contains_particle_ >> compositor_;
          contains_particle_ >> any_;
        }

        virtual Void
        traverse (SemanticGraph::Complex& c)
        {
          using SemanticGraph::Schema;

          traverse (static_cast<SemanticGraph::Type&> (c));

          //@@ Maybe use traverser?
          //
          SemanticGraph::Complex::InheritsIterator it (c.inherits_begin ());

          if (it != c.inherits_end ())
          {
            SemanticGraph::Type& t (it->inheritee ());

            if (t.named () &&
                types_.find (
                  t.scope ().name () + L"#" + t.name ()) == types_.end ())
            {
              // Don't worry about types that are in included/imported
              // schemas.
              //
              Schema& s (dynamic_cast<Schema&> (t.scope ().scope ()));

              if (&s == &schema || sources_p (schema, s))
              {
                valid = false;

                wcerr << c.file () << ":" << c.line () << ":" << c.column ()
                      << ": error: type '" << xpath (c) << "' inherits from "
                      << "yet undefined type '" << xpath (t) << "'" << endl;

                wcerr << t.file () << ":" << t.line () << ":" << t.column ()
                      << ": info: '" << xpath (t) << "' is defined here"
                      << endl;

                wcerr << c.file () << ":" << c.line () << ":" << c.column ()
                      << ": info: inheritance from a yet-undefined type is "
                      << "not supported" << endl;

                wcerr << c.file () << ":" << c.line () << ":" << c.column ()
                      << ": info: re-arrange your schema and try again"
                      << endl;
              }
            }
          }

          Traversal::Complex::traverse (c);
        }

        // Return true if root sources s.
        //
        Boolean
        sources_p (SemanticGraph::Schema& root, SemanticGraph::Schema& s)
        {
          using SemanticGraph::Schema;
          using SemanticGraph::Sources;

          for (Schema::UsesIterator i (root.uses_begin ());
               i != root.uses_end (); ++i)
          {
            if (i->is_a<Sources> ())
            {
              if (&i->schema () == &s || sources_p (i->schema (), s))
                return true;
            }
          }

          return false;
        }

        virtual Void
        traverse (SemanticGraph::Type& t)
        {
          if (t.named ())
          {
            types_.insert (t.scope ().name () + L"#" + t.name ());
          }
        }

        virtual Void
        traverse (SemanticGraph::Element& e)
        {
          /*
          SemanticGraph::Type& t (e.type ());

          // Check for forward inheritance.
          //
          if (dynamic_cast<SemanticGraph::AnyType*> (&t) != 0 &&
              dynamic_cast<SemanticGraph::Fundamental::Type*> (&t) == 0)
          {
            valid = false;

            wcerr << e.file () << ":" << e.line () << ":" << e.column ()
                  << ": error: element '" << xpath (e) << "' is of type "
                  << "'anyType'" << endl;

            wcerr << e.file () << ":" << e.line () << ":" << e.column ()
                  << ": info: 'anyType' is not supported" << endl;

            wcerr << e.file () << ":" << e.line () << ":" << e.column ()
                  << ": info: did you forget to specify 'type' attribute?"
                  << endl;
          }
          */
        }


        virtual Void
        traverse (SemanticGraph::Attribute& a)
        {
          /*
          SemanticGraph::Type& t (a.type ());

          if (dynamic_cast<SemanticGraph::AnySimpleType*> (&t) != 0 &&
              dynamic_cast<SemanticGraph::Fundamental::Type*> (&t) == 0)
          {
            valid = false;

            wcerr << a.file () << ":" << a.line () << ":" << a.column ()
                  << ": error: attribute '" << xpath (a) << "' is of type "
                  << "'anySimpleType'" << endl;

            wcerr << a.file () << ":" << a.line () << ":" << a.column ()
                  << ": info: 'anySimpleType' is not supported" << endl;

            wcerr << a.file () << ":" << a.line () << ":" << a.column ()
                  << ": info: did you forget to specify 'type' attribute?"
                  << endl;
          }
          */
        }

      private:
        Containers::Set<String> types_;

        Traversal::Sources sources_;

        Traversal::Names schema_names_;
        Traversal::Namespace ns_;

        Traversal::Names names_;

        // Any.
        //
        Any any_;
        Traversal::Compositor compositor_;
        Traversal::ContainsParticle contains_particle_;
        Traversal::ContainsCompositor contains_compositor_;
      };


      struct AnonymousType : Traversal::Schema,
                             Traversal::Complex,
                             Traversal::Element,
                             Traversal::Attribute,
                             protected virtual ValidationContext
      {
        AnonymousType (ValidationContext& c)
            : ValidationContext (c)
        {
          *this >> sources_ >> *this;

          *this >> schema_names_ >> ns_ >> names_ >> *this;
          *this >> names_;
        }

        virtual Void
        traverse (SemanticGraph::Element& e)
        {
	  if (skip (e)) return;

          SemanticGraph::Type& t (e.type ());

          if (!t.named ()
              && !t.is_a<SemanticGraph::Fundamental::IdRef> ()
              && !t.is_a<SemanticGraph::Fundamental::IdRefs> ())
          {
            valid = false;

            wcerr << e.file () << ":" << e.line () << ":" << e.column ()
                  << ": error: element '" << xpath (e) << "' "
                  << "is of anonymous type" << endl;

            wcerr << e.file () << ":" << e.line () << ":" << e.column ()
                  << ": info: "
                  << "anonymous types are not supported in this mapping"
                  << endl;

            wcerr << e.file () << ":" << e.line () << ":" << e.column ()
                  << ": info: "
                  << "use --morph-anonymous to automatically name "
                  << "anonymous types"
                  << endl;
          }
          else
          {
            Traversal::Element::traverse (e);
          }
        }

        virtual Void
        traverse (SemanticGraph::Attribute& a)
        {
          SemanticGraph::Type& t (a.type ());

          if (!t.named ()
              && !t.is_a<SemanticGraph::Fundamental::IdRef> ()
              && !t.is_a<SemanticGraph::Fundamental::IdRefs> ())
          {
            valid = false;

            wcerr << a.file () << ":" << a.line () << ":" << a.column ()
                  << ": error: attribute '" << xpath (a) << "' "
                  << "is of anonymous type" << endl;

            wcerr << a.file () << ":" << a.line () << ":" << a.column ()
                  << ": info: "
                  << "anonymous types are not supported in this mapping"
                  << endl;

            wcerr << a.file () << ":" << a.line () << ":" << a.column ()
                  << ": info: "
                  << "use --morph-anonymous to automatically name "
                  << "anonymous types"
                  << endl;
          }
          else
          {
            Traversal::Attribute::traverse (a);
          }
        }

      private:
        Containers::Set<String> types_;

        Traversal::Sources sources_;

        Traversal::Names schema_names_;
        Traversal::Namespace ns_;

        Traversal::Names names_;
      };
    }

    Validator::
    Validator ()
    {
      // Dummy ctor, help with long symbols on HP-UX.
    }

    Boolean Validator::
    validate (CLI::Options const& options,
              SemanticGraph::Schema& schema,
              SemanticGraph::Path const& tu)
    {
      //
      //
      if (options.value<CLI::char_type> () != "char" &&
          options.value<CLI::char_type> () != "wchar_t")
      {
        wcerr << "warning: unknown base character type '" <<
          options.value<CLI::char_type> ().c_str () << "'" << endl;
      }

      //
      //
      if (options.value<CLI::xml_parser> () != "xerces" &&
          options.value<CLI::xml_parser> () != "expat")
      {
        wcerr << "warning: unknown underlying XML parser '" <<
          options.value<CLI::xml_parser> ().c_str () << "'" << endl;
      }

      //
      //
      if (options.value<CLI::xml_parser> () == "expat" &&
          options.value<CLI::char_type> () == "wchar_t")
      {
        wcerr << "error: using expat with wchar_t is not yet supported"
              << endl;

        return false;
      }


      Boolean valid (true);
      ValidationContext ctx (schema, options, valid);

      // Test for anonymout types.
      //
      {
        AnonymousType traverser (ctx);
        traverser.dispatch (schema);
      }


      // Test the rest.
      //
      if (valid)
      {
        Traverser traverser (ctx);
        traverser.dispatch (schema);
      }

      return valid;
    }
  }
}

